/*
 * client.h: client structure declarations
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef IRCD_CLIENT_H
#define IRCD_CLIENT_H

struct client {
    char    nick[NICKLEN + 1];      /* nickname */
    char    user[USERLEN + 1];      /* username on IRC (different from username
                                       in conn */
    char    host[HOSTLEN + 1];      /* hostname on IRC */
    char    *orighost;              /* concession for host-changing modules.  this
                                       will normally point to the 'host' field of
                                       the client, but may be pointed elsewhere if
                                       need be. */
    char    ip[IPADDR_MAXLEN + 1];  /* IP address (NICKIP) */
    char    info[GCOSLEN + 1];      /* gecos/gcos info (ircname) */

#define IRCD_CLIENT_REGISTERED        0x0001
    /* check to see if the client is registered or not */
#define CLIENT_REGISTERED(cli) (cli->flags & IRCD_CLIENT_REGISTERED)
#define IRCD_CLIENT_KILLED        0x0002
#define IRCD_CLIENT_HISTORY        0x0004
    int     flags;

    /* check to see if this is our client.  we know it's ours if their server
     * points to us.  this was a check to see if conn was NULL, but conn can be
     * NULL for pseduo-clients. */
#define MYCLIENT(cli) (cli->server == ircd.me)
    struct connection *conn;        /* the connection for this client, NULL if
                                       remote or pseudo-client. */

    time_t  signon;                 /* signon time */
    time_t  ts;                     /* timestamp of the nickname */
    time_t  last;                   /* used for idle time (different from
                                       conn->last) */
    int     hops;                   /* how many hops away are they? */
    uint64_t modes;                 /* the user's modes, see below */

    struct userchans chans;         /* our channels */

    struct privilege_set *pset;     /* privileges (usually derived from class) */
    struct server *server;          /* server which owns this client */

    struct client_history *hist;

    char    *mdext;                 /* mdext data */
    
    LIST_ENTRY(client) lp;
};

struct client *create_client(struct connection *);
void destroy_client(struct client *, char *);
void client_change_nick(struct client *, char *);

/* find any client by name, including unregistered clients */
#define find_client_any(name) (client_t *)hash_find(ircd.hashes.client, name)
/* this is the most common case.  only find registered clients. */
client_t *find_client(char *);

/* this registers a client on the network/server.  registering a client
 * consists of checking it against registered acls, placing it in a connection
 * class (for real, not the default) (or dropping it if the class is full),
 * then propogating the user creation across the network, and finally
 * welcoming the user to IRC.  The function returns 1 if all this is
 * successful, 0 otherwise */
int register_client(struct client *);

/* check to see that a nickname is valid. */
int check_nickname(char *);

char *make_client_mask(char *mask);

TAILQ_HEAD(client_history_list, client_history);
struct client_history {
    char    nick[NICKLEN + 1];      /* these are all the same as in the */
    char    serv[SERVLEN + 1];

    client_t *cli;                  /* points to the client we created this
                                       from.  for currently online clients the
                                       nick will be different. */
    time_t  signoff;                /* when the client signed off */

    TAILQ_ENTRY(client_history) lp;
};

struct client_history *client_add_history(client_t *);
#define client_find_history(nick)                                             \
(struct client_history *)hash_find(ircd.hashes.client_history, (nick))
client_t *client_get_history(char *, time_t);
#define client_chase client_get_history

char **client_mdext_iter(char **);

/* this structure is used by client_check_access to pass to its various hook
 * functions when checking for access */
struct client_check_args {
    client_t *from;                 /* the client performing the action */
    client_t *to;                   /* the client being acted on */
    char    *extra;                 /* any extra data */
};

#define CLIENT_CHECK_OVERRIDE   HOOK_COND_SPASS
#define CLIENT_CHECK_OK         HOOK_COND_PASS
#define CLIENT_CHECK_NO         HOOK_COND_FAIL
int client_check_access(client_t *, client_t *, char *, event_t *);
#define can_can_send_client(from, to, arg) \
client_check_access(cli, to, arg, ircd.events.can_send_client)
#define can_can_nick_client(cli, arg) \
client_check_access(cli, NULL, arg, ircd.events.can_nick_client)

/* this is used by the hash function to compare nicks, use our resident
 * table */
int nickcmp(char *, char *, size_t);

#endif
/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
