/*
 * usermode.c: client (user) mode management functions
 * 
 * Copyright 2002-2007 the Ithildin Project.
 * See the COPYING file for more information on licensing and use.
 * 
 * This file contains the generic code for manipulating user modes, as well
 * as some base usermodes supported by the server.
 */

#include <ithildin/ithildin.h>

#include "ircd.h"

IDSTRING(rcsid, "$Id$");

static USERMODE_FUNC(mode_set_counter);

/* This function either initializes the usermode system or handles updates
 * in the case of a reload. */
void usermode_init(bool reload) {
    uint64_t ui64 = 1;
    unsigned char c;

#define MODELOOP(from, to) do {                                               \
    for (c = from;c <= to;c++, ui64 <<= 1) {                                  \
        ircd.umodes.modes[c].mode = c;                                        \
        ircd.umodes.modes[c].mask = ui64;                                     \
        ircd.umodes.modes[c].avail = 1;                                       \
    }                                                                         \
} while (0)

    if (reload == false) {
        MODELOOP('a', 'z'); /* set a-z modes */
        MODELOOP('A', 'Z'); /* and A-Z modes */
        MODELOOP('0', '9'); /* and 0-9 modes */

        /* we know that all modes are free.  grab i, o, and s */
        ircd.umodes.mode_invis = usermode_request('i', &c, USERMODE_FL_GLOBAL,
                -1, mode_set_counter);
        ircd.umodes.mode_oper = usermode_request('o', &c, USERMODE_FL_GLOBAL,
                ircd.sflag.ops, mode_set_counter);
        ircd.umodes.mode_servmsg = usermode_request('s', &c, 0,
                ircd.sflag.servmsg, NULL);
    } else {
        /* Be sure to update the function address for this channel mode in
         * case it moved.. */
        usermode_update_func(ircd.umodes.mode_invis, mode_set_counter);
        usermode_update_func(ircd.umodes.mode_oper, mode_set_counter);
        usermode_update_func(ircd.umodes.mode_servmsg, mode_set_counter);
    }
}

void usermode_deinit(bool reload) {
    if (reload == false) {
        /* dump our modes */
        usermode_release(ircd.umodes.mode_invis);
        usermode_release(ircd.umodes.mode_oper);
        usermode_release(ircd.umodes.mode_servmsg);
    }
}

uint64_t usermode_request(unsigned char suggested, unsigned char *actual,
        int flags, int sflag, usermode_func changefunc) {
    struct usermode *md = NULL;
    uint64_t allmodes = 0;
    int i;
    unsigned char c;

    /* build a list of all our currently in-use modes (some modes are
     * inaccessible, so only do it for modes which have been initialized and
     * are unavailable */
    for (i = 0;i < 256;i++) {
        if (!ircd.umodes.modes[i].avail && ircd.umodes.modes[i].mask)
            allmodes |= ircd.umodes.modes[i].mask;
    }

    if (ircd.umodes.modes[suggested].avail)
        md = &ircd.umodes.modes[suggested];
    else {
        if (islower(suggested))
            c = toupper(suggested);
        else
            c = tolower(suggested);

        if (ircd.umodes.modes[c].avail)
            md = &ircd.umodes.modes[c];
        else {
            for (i = 0;i < 256;i++) {
                if (ircd.umodes.modes[i].avail && ircd.umodes.modes[i].mask) {
                    md = &ircd.umodes.modes[i];
                    break;
                }
            }

            /* *still* NULL?  nothing free */
            if (md == NULL)
                return 0;
        }
    }

    md->avail = 0; /* mark it used */
    md->flags = flags;
    md->changefunc = changefunc;
    md->sflag = sflag;
    *actual = md->mode;

    allmodes |= md->mask;

    /* rebuild our mode string */
    strcpy(ircd.umodes.avail, usermode_getstr(allmodes, 0) + 1);
    
    return md->mask;
}

void usermode_release(unsigned char mode) {
    uint64_t allmodes = 0;
    int i;

    /* same as above, see what is available to rebuild the modestring */
    for (i = 0;i < 256;i++) {
        if (!ircd.umodes.modes[i].avail && ircd.umodes.modes[i].mask)
            allmodes |= ircd.umodes.modes[i].mask;
    }
    
    /* mark our released mode available, and remove it from the potential list
     * of modes */
    ircd.umodes.modes[mode].avail = 1;
    ircd.umodes.modes[mode].flags = 0;
    ircd.umodes.modes[mode].changefunc = NULL;
    allmodes &= ~ircd.umodes.modes[mode].mask;

    strcpy(ircd.umodes.avail, usermode_getstr(allmodes, 0) + 1);
}

void usermode_update_func(unsigned char mode, usermode_func changefunc) {
    struct usermode *md = &ircd.umodes.modes[mode];

    if (md->avail || !md->mask)
        return;

    md->changefunc = changefunc;
}

unsigned char *usermode_getstr(uint64_t modes, char global) {
    int i;
    int si; /* index to string */
    static unsigned char string[66];

    string[0] = '+'; /* prefix with a + */

    for (i = 0, si = 1;i < 256;i++) {
        if (modes & ircd.umodes.modes[i].mask &&
                (!global || ircd.umodes.modes[i].flags & USERMODE_FL_GLOBAL))
            string[si++] = ircd.umodes.modes[i].mode;
    }
    string[si] = '\0';

    return string;
}

uint64_t usermode_getmask(unsigned char *str, char global) {
    uint64_t mask = 0;
    unsigned char *s = str;

    while (*s) {
        if (!global || ircd.umodes.modes[*s].flags & USERMODE_FL_GLOBAL)
           mask |= ircd.umodes.modes[*s++].mask;
        else
            s++;
    }
    

    return mask;
}

void usermode_diff(uint64_t old, uint64_t new, char *result, char global) {
    int i;
    char *s = result;

    /* first see if they set anything */
    *s++ = '+';
    for (i = 0;i < 256;i++) {
        /* if this is a valid mode, see if it changed positively */
        if ((!global || ircd.umodes.modes[i].flags & USERMODE_FL_GLOBAL) &&
                !(old & ircd.umodes.modes[i].mask) && 
                (new & ircd.umodes.modes[i].mask))
            *s++ = ircd.umodes.modes[i].mode;
    }
    /* see if any changes were done positively, if not, dike the '+' */
    if (*(s - 1) == '+')
        s = s - 1;
    /* now see if they unset anything */
    *s++ = '-';
    for (i = 0;i < 256;i++) {
        /* if this is a valid mode, see if it changed negatively */
        if ((!global || ircd.umodes.modes[i].flags & USERMODE_FL_GLOBAL) &&
                (old & ircd.umodes.modes[i].mask) && 
                !(new & ircd.umodes.modes[i].mask))
            *s++ = ircd.umodes.modes[i].mode;
    }
    /* see if any changes were done negatively, if not, dike the '-' as done
     * above. */
    if (*(s - 1) == '-')
        s = s - 1;

    *s = '\0'; /* and terminate. */
}

int usermode_set(unsigned char mode, client_t *cli, client_t *on, char *arg,
        int *argused) {
    struct usermode *md = &ircd.umodes.modes[mode];

    if (md->avail || !md->mask)
        return 0;

    if (on->modes & md->mask)
        return 1; /* they already did it... */

    /* always call the set function.  if they're not our client, we ignore the
     * return value. */
    if ((md->changefunc != NULL &&
                !md->changefunc(cli, on, mode, 1, arg, argused)) && MYCLIENT(on))
        return 0;
    /* if they're local, not opered, and the mode is an oper mode, don't let
     * them set it. */
    if (MYCLIENT(on) && !CLIENT_OPERATOR(on) && md->flags & USERMODE_FL_OPER)
        return 0; /* no no. */

    /* okay, so, it's fine then, add away */
    on->modes |= md->mask;

    /* if the mode has a send flag, add them to the group for it (force the add
     * no matter what the settings for the sflag) */
    if (md->sflag > -1 && MYCLIENT(on))
        add_to_send_flag(md->sflag, on, true);

    return 1;
}


int usermode_unset(unsigned char mode, client_t *cli, client_t *on, char *arg,
        int *argused) {
    struct usermode *md = &ircd.umodes.modes[mode];

    if (md->avail || !md->mask)
        return 0;

    if (!(on->modes & md->mask))
        return 1; /* they're not set to that mode */

    /* always call the unset function.  if they're not our client, we ignore
     * the return value. */
    if ((md->changefunc != NULL &&
                !md->changefunc(cli, on, mode, 0, arg, argused)) && MYCLIENT(on))
        return 0;

    /* approved for removal */
    on->modes &= ~md->mask;

    /* if the mode has a send flag, remove them from the group for it */
    if (md->sflag > -1 && MYCLIENT(on))
        remove_from_send_flag(md->sflag, on, true);

    return 1;
}

/* ***************************************************************************
 * Code specific to individual usermodes
 * ***************************************************************************/
static USERMODE_FUNC(mode_set_counter) {
    int i;

    switch (mode) {
        case 'i':
            if (set)
                ircd.stats.net.visclients--;
            else
                ircd.stats.net.visclients++;
            break;
        case 'o':
            /* if this is our client, isn't set +o, and doesn't have the
             * privilege to do, deny. */
            if (MYCLIENT(cli) && by == cli && !CLIENT_OPERATOR(cli) &&
                    !BPRIV(cli, ircd.privileges.priv_operator))
                return 0;
            if (set) {
                ircd.stats.opers++;
                if (!MYCLIENT(cli))
                    /* change them to the default operator privilege set */
                    cli->pset = ircd.privileges.oper_set;
                hook_event(ircd.events.client_oper, cli);
            } else {
                /* remove them from any operator-only send flags, also remove
                 * any operator-only modes they might have set */
                if (MYCLIENT(cli)) {
                    unsigned char *s;
                    for (s = ircd.umodes.avail;*s != '\0';s++) {
                        if (ircd.umodes.modes[*s].flags & USERMODE_FL_OPER &&
                                !(ircd.umodes.modes[*s].flags &
                                    USERMODE_FL_PRESERVE))
                            usermode_unset(*s, cli, cli, NULL, NULL);
                    }
                    for (i = 0;i < ircd.sflag.size;i++) {
                        if (ircd.sflag.flags[i].flags &
                                SEND_LEVEL_OPERATOR &&
                            !(ircd.sflag.flags[i].flags &
                                SEND_LEVEL_PRESERVE))
                            remove_from_send_flag(i, cli, true);
                    }
                }
                ircd.stats.opers--;
                hook_event(ircd.events.client_deoper, cli);
            }
            break;
    }

    return 1;
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
