/*
 * superstring_private.h: Private support data for the SuperString library.
 * 
 * Copyright 2007 Chip Norkus
 * See the COPYING file for more information on licensing and use.
 * 
 * $Id$
 */

#ifndef SUPERSTRING_PRIVATE_H
#define SUPERSTRING_PRIVATE_H

#include <assert.h>
#include <string.h>

extern struct _superstring_global_data_ {
    superstring_malloc_function malloc;
    superstring_realloc_function realloc;
    superstring_free_function free;
} superstring_global_data;

#define ss_malloc(x) superstring_global_data.malloc((x));
#define ss_realloc(x, y) superstring_global_data.realloc((x), (y))
#define ss_free(x) superstring_global_data.free((x))

/* Internally we map all superstrings into this structure.  It mostly
 * resembles a non-debug jumbo superstring, but also points to the beginning
 * of data in the real string */
struct _superstring_internal_ {
    uint16_t flags;
    size_t len;
    size_t size;
    const unsigned char *data;
    superstring str;
};
typedef struct _superstring_internal_ ss_internal;

/* These two functions convert superstrings to ss_internal representations
 * and vice versa.  They are called when a function first begins work on the
 * string and at the end of a function when the string has been modified. */
static inline void ss_to_ssi(superstring str, ss_internal *istr) {
    istr->data = ss_cstring(str);
    if ((istr->flags = str->flags) & SUPERSTRING_ANONYMOUS) {
        istr->len = strlen(ss_cstring(str));
        istr->size = istr->len + 1;
        istr->str = NULL; /* we can't touch this, don't even try! */
    } else {
        istr->len = ss_length(str);
        istr->size = ss_size(str);
        istr->str = str;
    }
}

static inline void ssi_to_ss(ss_internal *istr, superstring str) {
    /* bail out if this is an anonymous value */
    if (istr->flags & SUPERSTRING_ANONYMOUS)
        return;

    /* might be jumbo */
    str->flags = istr->flags;
    if (istr->flags & SUPERSTRING_JUMBO) {
        struct _superstring_jumbo_ *jstr = (struct _superstring_jumbo_ *)str;
        jstr->len = istr->len;
        jstr->size = istr->size;
    } else {
        str->len = istr->len;
        str->size = istr->size;
    }
}

/* Push a C string into an anonymous internal superstring (with lies) */
static inline void cstr_to_ssi(const unsigned char *str, ss_internal *istr) {
    /* always make it jumbo for reasons of my sanity */
    istr->flags = SUPERSTRING_ANONYMOUS | SUPERSTRING_JUMBO;
    istr->data = str;
    istr->len = strlen(str);
    istr->size = istr->len + 1;
    istr->str = NULL;
}

/* three tables for ascii data */
extern const unsigned char ss_ascii_tab[256];
extern const unsigned char ss_ascii_lowertab[256];
extern const unsigned char ss_ascii_uppertab[256];

/* A macro to set ss_errno and return the error */
#define ss_returnerror(x) return (ss_errno = (x))

/* Function to grow strings */
int ss_grow(ss_internal *istr, superstring *outstr);

/* Handy stuff for handling printf-like functions and warnings in gcc (thanks
 * hop & larne @ EFnet) */
/* some useful gcc-only defines to tag purposefully unused variables, and to
 * tag functions as being printf-like */
#if defined(__GNUC__)
# define __UNUSED __attribute__ ((__unused__))
# define __PRINTF(x) __attribute__ ((format (printf, x, x + 1)))
# ifdef __FreeBSD__
#  define __PRINTF0(x) __attribute__ ((format (printf0, x, x + 1)))
# else
#  define __PRINTF0(x) __attribute__ ((format (printf, x, x + 1)))
# endif
#else
# define __UNUSED
# define __PRINTF(x)
# define __PRINTF0(x)
#endif
#define IDSTRING(var,string) static const char var[] __UNUSED = string

#endif /* SUPERSTRING_PRIVATE_H */

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
