/*
 * test.c: Crummy unit tests for SuperString library. :)
 * 
 * Copyright 2007 Chip Norkus.
 * See the COPYING file for more information on licensing and use.
 */

#include <superstring/superstring.h>
#include "superstring_private.h"
#include <stdio.h>

IDSTRING(rcsid, "$Id$");

#define runtest(name) if (true) {                                             \
    printf("Testing for " #name "\n");                                        \
    test_ ## name();                                                          \
} else (void *)0

#define tassert(expr) if (true) {                                             \
    if (!(expr)) {                                                            \
        printf("%s@%s:%d: \"%s\" failed.\n", __FUNCTION__, __FILE__,          \
               __LINE__, #expr);                                              \
        exit(1);                                                              \
    }                                                                         \
} else (void *)0

void test_allocator(void);
void test_allocator_setting(void);
void test_anonymous(void);
void test_compare(void);
void test_setchar(void);
void test_append(void);

#define TEST_INIT_VALUE "Hello, world!"
#define TEST_INIT_VALUE_LOWER "hello, world!"

void test_allocator(void) {
    superstring str, dstr;

    /* should be able to create with various types of bad data */
    tassert((str = ss_create(NULL)) != NULL);
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);
    tassert(str == NULL);
    tassert((str = ss_create("")) != NULL);
    tassert(ss_length(str) == 0);
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);
    tassert(str == NULL);

    /* Don't barf on simple initializations of course :) */
    str = ss_create(TEST_INIT_VALUE);
    tassert(str != NULL);
    tassert(strcmp(ss_cstring(str), TEST_INIT_VALUE) == 0);
    tassert(strlen(TEST_INIT_VALUE) == ss_length(str));
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);

    /* Ensure advisory size is ignored */
    str = ss_create_ex(TEST_INIT_VALUE, 0, 0, 1);
    tassert(str != NULL);
    tassert(ss_size(str) > 1);
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);

    /* Ensure advisory size is NOT ignored here */
    tassert((str = ss_create_ex(TEST_INIT_VALUE, SUPERSTRING_FIXED, 0, 1)) == NULL);
    tassert(ss_errno == SS_ERROR_INVALID_PARAMETER);

    /* Allocate a jumbo string and ensure it works correctly.  It should be
     * flagged jumbo and the length/size/cstring inlines should pull out the
     * correct data */
    tassert((str = ss_create_ex(TEST_INIT_VALUE, SUPERSTRING_JUMBO, 0, 1024)) != NULL);
    tassert(ss_size(str) == 1024);
    tassert(str->flags & SUPERSTRING_JUMBO);
    tassert(strcmp(ss_cstring(str), TEST_INIT_VALUE) == 0);
    tassert(strlen(TEST_INIT_VALUE) == ss_length(str));
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);

    /* Make sure that length is properly honored. */
    tassert((str = ss_create_ex(TEST_INIT_VALUE, 0, 5, 0)) != NULL);
    tassert(ss_length(str) == 5);
    tassert(strcmp(ss_cstring(str), TEST_INIT_VALUE) != 0);
    tassert(strncmp(ss_cstring(str), TEST_INIT_VALUE, 5) == 0);

    dstr = ss_duplicate(str);
    tassert(dstr != str);
    tassert(ss_size(dstr) == ss_size(str));
    tassert(ss_length(dstr) == ss_length(str));
    tassert(strcmp(ss_cstring(str), ss_cstring(dstr)) == 0);
    tassert(ss_destroy(&str) == SS_ERROR_SUCCESS);
}

void test_allocator_setting(void) {

    /* ensure that the allocator has been properly initialized and that we
     * cannot set bogus values on it */
    tassert(superstring_global_data.malloc == malloc);
    tassert(superstring_global_data.realloc == realloc);
    tassert(superstring_global_data.free == free);

    /* Make sure we can't do a dastardly deed.. */
    tassert(ss_allocator_set(malloc, NULL, free) != SS_ERROR_SUCCESS);
    tassert(ss_errno == SS_ERROR_INVALID_PARAMETER);
}

void test_anonymous(void) {
    superstring str;

    /* Make sure we can't do anything dumb with anonymous strings.  In
     * general anonymous strings are not suitable for much of anything.
     * length/size should be 0, they should not be able to be set writeable,
     * and deleting or appending to them is out of the question */
    str = SS_TEMP(TEST_INIT_VALUE);

    tassert(ss_readonly(str));
    tassert(ss_size(str) == 0 && ss_length(str) == 0);
    tassert(strcmp(ss_cstring(str), TEST_INIT_VALUE) == 0);
    tassert(ss_setwriteable(str) != SS_ERROR_SUCCESS);
    tassert(ss_destroy(&str) != SS_ERROR_SUCCESS);
}

void test_compare(void) {
    superstring lstr, rstr;

    /* easy check */
    lstr = ss_create(TEST_INIT_VALUE);
    rstr = ss_create(TEST_INIT_VALUE);
    tassert(ss_compare(lstr, rstr) == 0);

    /* play with values that are only different in case */
    ss_destroy(&rstr);
    rstr = ss_create(TEST_INIT_VALUE_LOWER);
    tassert(ss_compare_cs(lstr, TEST_INIT_VALUE) == 0);
    tassert(ss_compare(lstr, rstr) == -1);
    tassert(ss_compare_cs(lstr, TEST_INIT_VALUE_LOWER) == -1);
    tassert(ss_icompare(lstr, rstr) == 0);

    tassert(ss_nicompare(lstr, rstr, 1024) == 0);

    /* do something harder, these values are the same, then differ in case,
     * then just differ outright. */
    ss_destroy(&lstr);
    ss_destroy(&rstr);
    lstr = ss_create("bees are awesome");
    rstr = ss_create("Bees Are Killer");
    
    tassert(ss_compare(lstr, rstr) != 0);
    tassert(ss_icompare(lstr, rstr) != 0);
    tassert(ss_nicompare(lstr, rstr, 9) == 0);
    tassert(ss_nicompare(lstr, rstr, 10) != 0);

    /* Ensure no stupidity for null-values in created strings */
    ss_destroy(&lstr);
    ss_destroy(&rstr);
    lstr = ss_create_ex("bees\0are\0awesome", 0, 16, 0);
    rstr = ss_create_ex("BEES\0ARE\0KILLER", 0, 15, 0);
    tassert(ss_length(lstr) == 16);
    tassert(ss_length(rstr) == 15);
    tassert(ss_compare(lstr, rstr) != 0);
    tassert(ss_nicompare(lstr, rstr, 9) == 0);
}

void test_setchar(void) {
    superstring str = ss_create(TEST_INIT_VALUE);

    tassert(ss_char(str, 0) == TEST_INIT_VALUE[0]);
    tassert(ss_setchar(str, 0, 'y') == SS_ERROR_SUCCESS);
    tassert(ss_char(str, 0) == 'y');
    tassert(ss_char(str, 0) != TEST_INIT_VALUE[0]);

    tassert(ss_setchar(str, 1024, '!') == SS_ERROR_INVALID_PARAMETER);
}

void test_append(void) {
    superstring str = ss_create("abc");
    superstring str2 = ss_create("123");

    tassert(ss_append(&str, NULL) == SS_ERROR_INVALID_PARAMETER);
    tassert(ss_append(&str, str2) == SS_ERROR_SUCCESS);
    tassert(ss_compare_cs(str, "abc123") == 0);
    ss_destroy(&str);
    ss_destroy(&str2);

    /* test an allocation which forces growth */
    size_t len;
    str = ss_create("abc");
    len = ss_length(str);
    for (int i = 0;i < 100;i++) {
        tassert(ss_append_cs(&str, "123") == SS_ERROR_SUCCESS);
        len += 3; /* length should grow by 3 .. */
    }
    tassert(ss_length(str) == len);

    ss_destroy(&str);
}

int main(void) {

    runtest(allocator);
    runtest(allocator_setting);
    runtest(anonymous);
    runtest(compare);
    runtest(setchar);
    runtest(append);

    /* If we made it here we are at the end!  Let them know we succeeded. */
    printf("All SuperString tests passed.\n");
    return 0;
}

/* vi:set ts=8 sts=4 sw=4 tw=76 et: */
